/*
 *  $Id$
 *
 * RTEMS Project (http://www.rtems.org/)
 *
 * Copyright 2007 Chris Johns (chrisj@rtems.org)
 */

/**
 * Basic File System test. This is to just check a file system
 * down work as expected.
 *
 * Please add more tests
 */

#include <errno.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <rtems.h>

#include "fs-test.h"

/**
 * Define bool if it does not exist.
 */
#if !defined (bool)
#define bool int
#endif

/**
 * Define true if it does not exist.
 */
#if !defined (true)
#define true (1)
#endif

/**
 * Define true if it does not exist.
 */
#if !defined (false)
#define false (0)
#endif

/**
 * The block size of files when creating.
 */
#define FILE_BLOCK_SIZE (256)

/**
 * The types for the file-system test.
 */
typedef enum fst_type
{
  fst_file,
  fst_dir
} fst_type;

/**
 * Jump to the top on error.
 */
static jmp_buf top;

/**
 * Stat a path returning the stat buffer.
 *
 * @param path Path to get the stat buffer of.
 * @param buf The stat buffer.
 */
static void
checked_stat (const char* path, struct stat* buf)
{
  if (stat (path, buf) < 0)
  {
    printf ("error: stat of '%s' failed: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Return if the path points to a directory.
 *
 * @param path Path to check if a directory.
 */
static void
check_dir (const char* path)
{
  struct stat ds;
  checked_stat (path, &ds);
  if (!S_ISDIR(ds.st_mode))
  {
    printf ("error: path '%s' is not a directory\n", path);
    longjmp (top, 1);
  }
}

/**
 * Return if the change to the path is successful.
 *
 * @param path Path to change directory to.
 */
static void
change_dir (const char* path)
{
  check_dir (path);
  if (chdir (path) < 0)
  {
    printf ("error: change directory to '%s' failed: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Return if the directory is created. The path is checked
 * to insure the full path exists and if it does not all parts
 * are made.
 *
 * @param dirpath Path to the directory to create.
 */
static bool
create_dir (const char* dirpath, bool jump_on_error)
{
  /*
   * The local copy can be written to.
   */
  char* path;
  char* next;

  /*
   * If the directory path is absolute take that as the full path
   * to create. If the path is relative get the current working
   * directory and pre-pend it to th directoy path.
   */

  if (*dirpath == '/')
  {
    path = strdup (dirpath);
    if (!path)
    {
      printf ("error: could not duplicate path in create_dir\n");
      longjmp (top, 1);
    }
  }
  else
  {
    /*
     * Find a buffer large enough for the current working directory.
     */
    size_t size = 256;
    path = NULL;
    while (path == NULL)
    {
      char* cwd = malloc (size);
      if (!cwd)
      {
        printf ("error: could not allocate memory in create_dir\n");
        longjmp (top, 1);
      }
      
      if (getcwd (cwd, size) == cwd)
      {
        path = malloc (strlen (cwd) + strlen (dirpath) + 2);

        if (!path)
        {
          free (cwd);
          printf ("error: could not duplicate path in create_dir\n");
          longjmp (top, 1);
        }

        strcpy (path, cwd);
        strcat (path, "/"); 
        strcpy (path, dirpath);

        free (cwd);
        break;
      }

      free (cwd);
      size += 256;
    }
  }
  
  next = path;

  /*
   * Break the path up into its parts and check to see if
   * each part exists. It a part does not exist create it.
   * If the part is not a directory report as an error.
   */
  
  while (next)
  {
    char separator;
    struct stat ds;

    /*
     * Remove the separator and replace with a nul. If
     * no separator is found this is the end of the path.
     */
    
    next = strchr (next, '/');
    
    if (next)
    {
      separator = *next;
      *next = '\0';
    }

    if (stat (path, &ds) == 0)
    {
      if (!S_ISDIR(ds.st_mode))
      {
        printf ("error: path exists and is not a directory: %s\n", path);
        free (path);
        longjmp (top, 1);
      }
    }
    else if (errno != ENOENT)
    {
      printf ("error: path stat failure: %s: %s (%d)\n",
              path, strerror (errno), errno);
      free (path);
      longjmp (top, 1);
    }
    else
    {
      if (mkdir (path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
      {
        free (path);
        if (jump_on_error)
        {
          printf ("error: make directory '%s' failed: %s (%d)\n",
                  path, strerror (errno), errno);
          longjmp (top, 1);
        }
        return false;
      }
    }

    if (next)
    {
      *next = separator;
      next++;
    }
  }

  free (path);

  return true;
}

/**
 * Return if the file is created. The path is checked
 * to insure the full path exists and if it does not all parts
 * are made. The file is fill with repeating numbers.
 *
 * @param filepath Path to the file to create.
 */
static bool
create_file (const char* filepath, size_t size, bool jump_on_error)
{
  unsigned char* block;
  unsigned char  pattern = 0;
  int fd;
  
  /*
   * Does the file path contain directories ?
   */
  if (strchr (filepath, '/') != NULL)
  {
    /*
     * Get a copy of the path and search backwards for the
     * directory separator.
     */
    
    char* dirpath = strdup (filepath);
    char* end = dirpath + strlen (dirpath);

    while (*end != '/')
      end--;

    *end = '\0';

    create_dir (dirpath, true);

    free (dirpath);
  }

  fd = open (filepath,
             O_CREAT | O_TRUNC | O_WRONLY,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  if (fd < 0)
  {
    if (jump_on_error)
    {
      printf ("error: open failed: %s: %s (%d)\n", filepath,
              strerror (errno), errno);
      longjmp (top, 1);
    }
    return false;
  }

  block = malloc (FILE_BLOCK_SIZE);

  if (!block)
  {
    printf ("error: no memory in create_file\n");
    close (fd);
    longjmp (top, 1);
  }
  
  while (size)
  {
    size_t bsize = size < FILE_BLOCK_SIZE ? size : FILE_BLOCK_SIZE;

    memset (block, pattern++, FILE_BLOCK_SIZE);

    if (write (fd, block, bsize) < 0)
    {
      printf ("error: file write failed: %s: %s (%d)\n",
              filepath, strerror (errno), errno);
      free (block);
      close (fd);
      longjmp (top, 1);
    }

    size -= bsize;
  }

  free (block);
  
  if (close (fd) < 0)
  {
    printf ("error: file close failed: %s: %s (%d)\n",
            filepath, strerror (errno), errno);
    longjmp (top, 1);
  }

  return true;
}

/**
 * Remove directory.
 *
 * @param path The path to the directory to remove.
 */
void
remove_dir (const char* path)
{
  if (rmdir (path) < 0)
  {
    printf ("error: removing directory: %s: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Remove file.
 *
 * @param path The path to the file to remove.
 */
void
remove_file (const char* path)
{
  if (unlink (path) < 0)
  {
    printf ("error: removing file: %s: %s (%d)\n",
            path, strerror (errno), errno);
    longjmp (top, 1);
  }
}

/**
 * Check the limit of a directory by creating directories or files
 * up to the limit.
 *
 * @param base 
 * @param type 
 * @param limit 
 */
void
check_dir_limit (const char* base, fst_type type, int limit)
{
  int count = 0;
  int c;
  
  create_dir (base, true);
  change_dir (base);
  
  while (count < limit)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", count);
    
    if (type == fst_file)
      create_file (path, 64, true);
    else
      create_dir (path, true);

    count++;
  }

  for (c = 0; c < count; c++)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", c);

    if (type == fst_file)
      remove_file (path);
    else
      remove_dir (path);
  }   
}

/**
 * Find the limit of a directory by creating directories or files
 * until an error is detected.
 *
 * @param base 
 * @param type 
 * @param limit 
 */
int
find_dir_limit (const char* base, fst_type type, int limit)
{
  int count = 0;
  int c;
  
  create_dir (base, true);
  change_dir (base);
  
  while (count < limit)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", count);
    
    if (type == fst_file)
    {
      if (!create_file (path, 64, false))
        break;
    }
    else
    {
      if (!create_dir (path, false))
        break;
    }

    count++;
  }

  for (c = 0; c < count; c++)
  {
    char path[16];

    snprintf (path, sizeof (path), "tst%04d", c);

    if (type == fst_file)
      remove_file (path);
    else
      remove_dir (path);
  }

  return count;
}

/**
 * Test the file system.
 */
void
filesystem_test (const char* base)
{
  int limit;
  
  printf ("File System test: %s\n", base);

/*
   * Catch any error and return.
   */
  if (setjmp (top))
    return;

  /*
   * Find out how many files will fit in the root directory.
   */
  printf ("Find root directory limits: files\n");
  limit = find_dir_limit (base, fst_file, 1000);
  /*
   * Make sure we can fill a subdirectory with more.
   */
  printf ("Checking directory limits: files\n");
  check_dir_limit (base, fst_file, limit);

  /*
   * Repeat with directories.
   */
  printf ("Checking directory limits: dir\n");
  check_dir_limit (base, fst_dir, limit);

  printf ("File System test: passed\n");
}
